#include "Material.h"
#include "Texture.h"
#include "RenderState.h"
#include "ShaderProgram.h"
#include "ShaderManager.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>


#ifdef _CRT_SET
	#define _CRTDBG_MAP_ALLOC
	#include <stdlib.h>
	#include <crtdbg.h>

	// redefine new operator
	#define DEBUG_NEW new(_NORMAL_BLOCK, __FILE__, __LINE__)
	#define new DEBUG_NEW
#endif


using namespace std;


namespace CHCDemoEngine
{


/////////
//-- cg stuff

static CGcontext sCgContext = NULL;

static CGprofile sCgFragmentProfile;
static CGprofile sCgVertexProfile; 

static ShaderProgram *sTreeAnimationProgramMrt;
static ShaderProgram *sTreeAnimationProgram;
static ShaderProgram *sTreeAnimation;

// only instance of the shader manager
ShaderManager *ShaderManager::sShaderManager = NULL;


static void cgErrorCallback()
{
	CGerror lastError = cgGetError();

	if(lastError)
	{
		printf("%s\n\n", cgGetErrorString(lastError));
		printf("%s\n", cgGetLastListing(sCgContext));
		
		printf("Cg error, exiting...\n");

		exit(0);
	}
}


ShaderManager::ShaderManager() 
{
	// setup cg
	cgSetErrorCallback(cgErrorCallback);
	// create context.
	sCgContext = cgCreateContext();
	// get the optimal fragment profile
	sCgFragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
	cgGLSetOptimalOptions(sCgFragmentProfile);
	// get the optimal vertex profile
	sCgVertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
	cgGLSetOptimalOptions(sCgVertexProfile);

	// set textures to auto load
	cgGLSetManageTextureParameters(sCgContext, false);

	mMrtDefaultVertexProgram = CreateVertexProgram("mrt", "vtx", "defaultMrtVertex");
	mMrtDefaultFragmentProgram = CreateFragmentProgram("mrt", "frag", "defaultMrtFragment");
	mMrtDefaultFragmentTexProgram = CreateFragmentProgram("mrt", "fragtex", "defaultMrtFragmentTex");

	// provide the current view matrix
	mMrtDefaultFragmentProgram->AddParameter("viewMatrix", 0);
	mMrtDefaultFragmentTexProgram->AddParameter("viewMatrix", 0);
	// add a texture parameter
	mMrtDefaultFragmentTexProgram->AddParameter("tex", 1);


	//////////////////////////
	//-- hack: apply hardcoded tree animation (should be provided with a material script!)

	sTreeAnimationProgram = CreateVertexProgram("treeanimation", "animateVtx", "treeAnimation");
	sTreeAnimationProgramMrt = CreateVertexProgram("treeanimation", "animateVtxMrt", "treeAnimationMrt");

	sTreeAnimationProgram->AddParameter("timer", 0);
	sTreeAnimationProgram->AddParameter("windDir", 1);
	sTreeAnimationProgram->AddParameter("windStrength", 2);
	sTreeAnimationProgram->AddParameter("minMaxPos", 3);
	sTreeAnimationProgram->AddParameter("frequency", 4);
	sTreeAnimationProgram->AddParameter("lightDir", 5);

	sTreeAnimationProgramMrt->AddParameter("timer", 0);
	sTreeAnimationProgramMrt->AddParameter("windDir", 1);
	sTreeAnimationProgramMrt->AddParameter("windStrength", 2);
	sTreeAnimationProgramMrt->AddParameter("minMaxPos", 3);
	sTreeAnimationProgramMrt->AddParameter("frequency", 4);

	cout << "cg initialization successful" << endl;
}


void ShaderManager::DelSingleton()
{
	DEL_PTR(sShaderManager);
}
	

ShaderManager::~ShaderManager()
{
	CLEAR_CONTAINER(mShaders);
	
	if (sCgContext)
		cgDestroyContext(sCgContext);
}


ShaderProgram *ShaderManager::CreateFragmentProgram(const std::string &filename, 
													const std::string &funcName,
													const std::string &name)
{
	return CreateShaderProgram(filename, funcName, name, true);
}


ShaderProgram *ShaderManager::CreateVertexProgram(const std::string &filename,
												  const std::string &funcName,
												  const std::string &name)
{
	return CreateShaderProgram(filename, funcName, name, false);
}


void ShaderManager::EnableFragmentProfile()
{
	cgGLEnableProfile(sCgFragmentProfile);
}


void ShaderManager::EnableVertexProfile()
{
	cgGLEnableProfile(sCgVertexProfile);
}


void ShaderManager::DisableFragmentProfile()
{
	cgGLDisableProfile(sCgFragmentProfile);
}


void ShaderManager::DisableVertexProfile()
{
	cgGLDisableProfile(sCgVertexProfile);
}


ShaderProgram *ShaderManager::CreateShaderProgram(const std::string &filename,
												  const std::string &funcName,
												  const std::string &name,
												  bool isFragment)
{
	const string fullName = "src/shaders/" + filename + ".cg";
	
	ShaderProgram *p = 
		new ShaderProgram(sCgContext, fullName, 
		                  isFragment ? sCgFragmentProfile : sCgVertexProfile, funcName);

	ShaderMap::const_iterator it = mShaderMap.find(name);

	if (it == mShaderMap.end())
	{
		DEL_PTR(p);

		cerr << "Program name " << name << "already chosen" << endl;
		return p;
	}
	
	if (!p->IsValid())
	{
		DEL_PTR(p);
		cerr << "Program " << funcName << " in " << fullName << " failed to load" << endl;
		return p;
	}

	cout << "Program " << funcName << " in " << fullName << " loaded" << endl;
		
	mShaders.push_back(p);
	mShaderMap[name] = p;

	return p;
}


}

